@@ -18,6 +18,7 @@ import time
 import attr
 from mozbuild.util import memoize
 from taskcluster.utils import fromNow
+from taskgraph import MAX_DEPENDENCIES
 from taskgraph.transforms.base import TransformSequence
 from taskgraph.util.copy import deepcopy
 from taskgraph.util.keyed_by import evaluate_keyed_by
@@ -31,11 +32,12 @@ from taskgraph.util.schema import (
 from taskgraph.util.treeherder import split_symbol
 from voluptuous import All, Any, Extra, Match, NotIn, Optional, Required
 
-from gecko_taskgraph import GECKO, MAX_DEPENDENCIES
+from gecko_taskgraph import GECKO
 from gecko_taskgraph.optimize.schema import OptimizationSchema
 from gecko_taskgraph.transforms.job.common import get_expiration
 from gecko_taskgraph.util import docker as dockerutil
 from gecko_taskgraph.util.attributes import TRUNK_PROJECTS, is_try, release_level
+from gecko_taskgraph.util.chunking import TEST_VARIANTS
 from gecko_taskgraph.util.hash import hash_path
 from gecko_taskgraph.util.partners import get_partners_to_be_published
 from gecko_taskgraph.util.scriptworker import BALROG_ACTIONS, get_release_config
@@ -146,8 +148,8 @@ task_description_schema = Schema(
             # 'by-tier' behavior will be used.
             "rank": Any(
                 # Rank is equal the timestamp of the build_date for tier-1
-                # tasks, and zero for non-tier-1.  This sorts tier-{2,3}
-                # builds below tier-1 in the index.
+                # tasks, and one for non-tier-1.  This sorts tier-{2,3}
+                # builds below tier-1 in the index, but above eager-index.
                 "by-tier",
                 # Rank is given as an integer constant (e.g. zero to make
                 # sure a task is last in the index).
@@ -198,6 +200,8 @@ task_description_schema = Schema(
         },
         # Override the default priority for the project
         Optional("priority"): str,
+        # Override the default 5 retries
+        Optional("retries"): int,
     }
 )
 
@@ -217,6 +221,7 @@ V2_ROUTE_TEMPLATES = [
     "index.{trust-domain}.v2.{project}.pushdate.{build_date}.latest.{product}.{job-name}",
     "index.{trust-domain}.v2.{project}.pushlog-id.{pushlog_id}.{product}.{job-name}",
     "index.{trust-domain}.v2.{project}.revision.{branch_rev}.{product}.{job-name}",
+    "index.{trust-domain}.v2.{project}.revision.{branch_git_rev}.{product}.{job-name}",
 ]
 
 # {central, inbound, autoland} write to a "trunk" index prefix. This facilitates
@@ -230,6 +235,7 @@ V2_SHIPPABLE_TEMPLATES = [
     "index.{trust-domain}.v2.{project}.shippable.{build_date}.revision.{branch_rev}.{product}.{job-name}",  # noqa - too long
     "index.{trust-domain}.v2.{project}.shippable.{build_date}.latest.{product}.{job-name}",
     "index.{trust-domain}.v2.{project}.shippable.revision.{branch_rev}.{product}.{job-name}",
+    "index.{trust-domain}.v2.{project}.shippable.revision.{branch_git_rev}.{product}.{job-name}",
 ]
 
 V2_SHIPPABLE_L10N_TEMPLATES = [
@@ -260,6 +266,12 @@ def get_branch_rev(config):
     ]
 
 
+def get_branch_git_rev(config):
+    return config.params[
+        "{}head_git_rev".format(config.graph_config["project-repo-param-prefix"])
+    ]
+
+
 def get_branch_repo(config):
     return config.params[
         "{}head_repository".format(
@@ -628,7 +640,9 @@ def build_docker_worker_payload(config, task, task_def):
 @payload_builder(
     "generic-worker",
     schema={
-        Required("os"): Any("windows", "macosx", "linux", "linux-bitbar"),
+        Required("os"): Any(
+            "windows", "macosx", "linux", "linux-bitbar", "linux-lambda"
+        ),
         # see http://schemas.taskcluster.net/generic-worker/v1/payload.json
         # and https://docs.taskcluster.net/reference/workers/generic-worker/payload
         # command is a list of commands to run, sequentially
@@ -683,7 +697,7 @@ def build_docker_worker_payload(config, task, task_def):
                 # Required if and only if `content` is specified and mounting a
                 # directory (not a file). This should be the archive format of the
                 # content (either pre-loaded cache or read-only directory).
-                Optional("format"): Any("rar", "tar.bz2", "tar.gz", "zip"),
+                Optional("format"): Any("rar", "tar.bz2", "tar.gz", "zip", "tar.xz"),
             }
         ],
         # environment variables
@@ -725,7 +739,7 @@ def build_generic_worker_payload(config, task, task_def):
         task_def["payload"].setdefault("onExitStatus", {}).setdefault(
             "retry", []
         ).extend(worker["retry-exit-status"])
-    if worker["os"] == "linux-bitbar":
+    if worker["os"] in ["linux-bitbar", "linux-lambda"]:
         task_def["payload"].setdefault("onExitStatus", {}).setdefault("retry", [])
         # exit code 4 is used to indicate an intermittent android device error
         if 4 not in task_def["payload"]["onExitStatus"]["retry"]:
@@ -1267,6 +1281,23 @@ def build_push_msix_payload(config, task, task_def):
         task_def["payload"]["publishMode"] = worker["publish-mode"]
 
 
+@payload_builder(
+    "shipit-update-product-channel-version",
+    schema={
+        Required("product"): str,
+        Required("channel"): str,
+        Required("version"): str,
+    },
+)
+def build_ship_it_update_product_channel_version_payload(config, task, task_def):
+    worker = task["worker"]
+    task_def["payload"] = {
+        "product": worker["product"],
+        "version": worker["version"],
+        "channel": worker["channel"],
+    }
+
+
 @payload_builder(
     "shipit-shipped",
     schema={
@@ -1499,6 +1530,177 @@ def build_treescript_payload(config, task, task_def):
         task_def["payload"]["ssh_user"] = worker["ssh-user"]
 
 
+@payload_builder(
+    "landoscript",
+    schema={
+        Required("lando-repo"): str,
+        Optional("hg-repo-url"): str,
+        Optional("ignore-closed-tree"): bool,
+        Optional("dontbuild"): bool,
+        Optional("tags"): [Any("buildN", "release", None)],
+        Optional("force-dry-run"): bool,
+        Optional("android-l10n-import-info"): {
+            Required("from-repo-url"): str,
+            Required("toml-info"): [
+                {
+                    Required("toml-path"): str,
+                    Required("dest-path"): str,
+                }
+            ],
+        },
+        Optional("android-l10n-sync-info"): {
+            Required("from-branch"): str,
+            Required("toml-info"): [
+                {
+                    Required("toml-path"): str,
+                }
+            ],
+        },
+        Optional("l10n-bump-info"): [
+            {
+                Required("name"): str,
+                Required("path"): str,
+                Optional("l10n-repo-url"): str,
+                Optional("l10n-repo-target-branch"): str,
+                Optional("ignore-config"): object,
+                Required("platform-configs"): [
+                    {
+                        Required("platforms"): [str],
+                        Required("path"): str,
+                        Optional("format"): str,
+                    }
+                ],
+            }
+        ],
+        Optional("bump-files"): [str],
+        Optional("merge-info"): object,
+    },
+)
+def build_landoscript_payload(config, task, task_def):
+    worker = task["worker"]
+    release_config = get_release_config(config)
+    task_def["payload"] = {"actions": [], "lando_repo": worker["lando-repo"]}
+    actions = task_def["payload"]["actions"]
+
+    if worker.get("ignore-closed-tree") is not None:
+        task_def["payload"]["ignore_closed_tree"] = worker["ignore-closed-tree"]
+
+    if worker.get("dontbuild"):
+        task_def["payload"]["dontbuild"] = True
+
+    if worker.get("force-dry-run"):
+        task_def["payload"]["dry_run"] = True
+
+    if worker.get("android-l10n-import-info"):
+        android_l10n_import_info = {}
+        for k, v in worker["android-l10n-import-info"].items():
+            android_l10n_import_info[k.replace("-", "_")] = worker[
+                "android-l10n-import-info"
+            ][k]
+        android_l10n_import_info["toml_info"] = [
+            {
+                param_name.replace("-", "_"): param_value
+                for param_name, param_value in entry.items()
+            }
+            for entry in worker["android-l10n-import-info"]["toml-info"]
+        ]
+        task_def["payload"]["android_l10n_import_info"] = android_l10n_import_info
+        actions.append("android_l10n_import")
+
+    if worker.get("android-l10n-sync-info"):
+        android_l10n_sync_info = {}
+        for k, v in worker["android-l10n-sync-info"].items():
+            android_l10n_sync_info[k.replace("-", "_")] = worker[
+                "android-l10n-sync-info"
+            ][k]
+        android_l10n_sync_info["toml_info"] = [
+            {
+                param_name.replace("-", "_"): param_value
+                for param_name, param_value in entry.items()
+            }
+            for entry in worker["android-l10n-sync-info"]["toml-info"]
+        ]
+        task_def["payload"]["android_l10n_sync_info"] = android_l10n_sync_info
+        actions.append("android_l10n_sync")
+
+    if worker.get("l10n-bump-info"):
+        l10n_bump_info = []
+        l10n_repo_urls = set()
+        for lbi in worker["l10n-bump-info"]:
+            new_lbi = {}
+            if "l10n-repo-url" in lbi:
+                l10n_repo_urls.add(lbi["l10n-repo-url"])
+            for k, v in lbi.items():
+                new_lbi[k.replace("-", "_")] = lbi[k]
+            l10n_bump_info.append(new_lbi)
+
+        task_def["payload"]["l10n_bump_info"] = l10n_bump_info
+        if len(l10n_repo_urls) > 1:
+            raise Exception(
+                "Must use the same l10n-repo-url for all files in the same task!"
+            )
+        elif len(l10n_repo_urls) == 1:
+            actions.append("l10n_bump")
+
+    if worker.get("tags"):
+        tag_names = []
+        product = task["shipping-product"].upper()
+        version = release_config["version"].replace(".", "_")
+        buildnum = release_config["build_number"]
+        if "buildN" in worker["tags"]:
+            tag_names.extend(
+                [
+                    f"{product}_{version}_BUILD{buildnum}",
+                ]
+            )
+        if "release" in worker["tags"]:
+            tag_names.extend([f"{product}_{version}_RELEASE"])
+        tag_info = {
+            "tags": tag_names,
+            "hg_repo_url": worker["hg-repo-url"],
+            "revision": config.params[
+                "{}head_rev".format(worker.get("repo-param-prefix", ""))
+            ],
+        }
+        task_def["payload"]["tag_info"] = tag_info
+        actions.append("tag")
+
+    if worker.get("bump-files"):
+        bump_info = {}
+        bump_info["next_version"] = release_config["next_version"]
+        bump_info["files"] = worker["bump-files"]
+        task_def["payload"]["version_bump_info"] = bump_info
+        actions.append("version_bump")
+
+    if worker.get("merge-info"):
+        merge_info = {
+            merge_param_name.replace("-", "_"): merge_param_value
+            for merge_param_name, merge_param_value in worker["merge-info"].items()
+            if merge_param_name != "version-files"
+        }
+        merge_info["version_files"] = [
+            {
+                file_param_name.replace("-", "_"): file_param_value
+                for file_param_name, file_param_value in file_entry.items()
+            }
+            for file_entry in worker["merge-info"]["version-files"]
+        ]
+        # hack alert: co-opt the l10n_bump_info into the merge_info section
+        # this should be cleaned up to avoid l10n_bump_info ever existing
+        # in the payload
+        if task_def["payload"].get("l10n_bump_info"):
+            actions.remove("l10n_bump")
+            merge_info["l10n_bump_info"] = task_def["payload"].pop("l10n_bump_info")
+
+        task_def["payload"]["merge_info"] = merge_info
+        actions.append("merge_day")
+
+    scopes = set(task_def.get("scopes", []))
+    scopes.add(f"project:releng:lando:repo:{worker['lando-repo']}")
+    scopes.update([f"project:releng:lando:action:{action}" for action in actions])
+    task_def["scopes"] = sorted(scopes)
+
+
 @payload_builder(
     "invalid",
     schema={
@@ -1531,27 +1733,46 @@ def set_implementation(config, tasks):
     Set the worker implementation based on the worker-type alias.
     """
     for task in tasks:
-        worker = task.setdefault("worker", {})
-        if "implementation" in task["worker"]:
-            yield task
-            continue
-
-        impl, os = worker_type_implementation(
+        default_worker_implementation, default_os = worker_type_implementation(
             config.graph_config, config.params, task["worker-type"]
         )
 
+        worker = task.setdefault("worker", {})
         tags = task.setdefault("tags", {})
-        tags["worker-implementation"] = impl
+
+        worker_implementation = worker.get(
+            "implementation", default_worker_implementation
+        )
+        tag_worker_implementation = _get_worker_implementation_tag(
+            config, task["worker-type"], worker_implementation
+        )
+        if worker_implementation:
+            worker["implementation"] = worker_implementation
+            tags["worker-implementation"] = tag_worker_implementation
+
+        os = worker.get("os", default_os)
         if os:
             tags["os"] = os
-
-        worker["implementation"] = impl
-        if os:
             worker["os"] = os
 
         yield task
 
 
+def _get_worker_implementation_tag(config, task_worker_type, worker_implementation):
+    # Scriptworkers have different types of payload and each sets its own
+    # worker-implementation. Per bug 1955941, we want to bundle them all in one category
+    # through their tags.
+    provisioner_id, _ = get_worker_type(
+        config.graph_config,
+        config.params,
+        task_worker_type,
+    )
+    if provisioner_id in ("scriptworker-k8s", "scriptworker-prov-v1"):
+        return "scriptworker"
+
+    return worker_implementation
+
+
 @transforms.add
 def set_defaults(config, tasks):
     for task in tasks:
@@ -1578,10 +1799,13 @@ def set_defaults(config, tasks):
         elif worker["implementation"] == "generic-worker":
             worker.setdefault("env", {})
             worker.setdefault("os-groups", [])
-            if worker["os-groups"] and worker["os"] != "windows":
+            if worker["os-groups"] and worker["os"] not in (
+                "windows",
+                "linux",
+            ):
                 raise Exception(
                     "os-groups feature of generic-worker is only supported on "
-                    "Windows, not on {}".format(worker["os"])
+                    "Windows and Linux, not on {}".format(worker["os"])
                 )
             worker.setdefault("chain-of-trust", False)
         elif worker["implementation"] in (
@@ -1625,7 +1849,7 @@ def task_name_from_label(config, tasks):
         if "label" not in task:
             if taskname is None:
                 raise Exception("task has neither a name nor a label")
-            task["label"] = "{}-{}".format(config.kind, taskname)
+            task["label"] = f"{config.kind}-{taskname}"
         yield task
 
 
@@ -1676,11 +1900,19 @@ def add_generic_index_routes(config, task):
     subs["product"] = index["product"]
     subs["trust-domain"] = config.graph_config["trust-domain"]
     subs["branch_rev"] = get_branch_rev(config)
+    try:
+        subs["branch_git_rev"] = get_branch_git_rev(config)
+    except KeyError:
+        pass
 
     project = config.params.get("project")
 
     for tpl in V2_ROUTE_TEMPLATES:
-        routes.append(tpl.format(**subs))
+        try:
+            routes.append(tpl.format(**subs))
+        except KeyError:
+            # Ignore errors that arise from branch_git_rev not being set.
+            pass
 
     # Additionally alias all tasks for "trunk" repos into a common
     # namespace.
@@ -1709,9 +1941,17 @@ def add_shippable_index_routes(config, task):
     subs["product"] = index["product"]
     subs["trust-domain"] = config.graph_config["trust-domain"]
     subs["branch_rev"] = get_branch_rev(config)
+    try:
+        subs["branch_git_rev"] = get_branch_git_rev(config)
+    except KeyError:
+        pass
 
     for tpl in V2_SHIPPABLE_TEMPLATES:
-        routes.append(tpl.format(**subs))
+        try:
+            routes.append(tpl.format(**subs))
+        except KeyError:
+            # Ignore errors that arise from branch_git_rev not being set.
+            pass
 
     # Also add routes for en-US
     task = add_shippable_l10n_index_routes(config, task, force_locale="en-US")
@@ -1855,10 +2095,11 @@ def add_index_routes(config, tasks):
         rank = index.get("rank", "by-tier")
 
         if rank == "by-tier":
-            # rank is zero for non-tier-1 tasks and based on pushid for others;
-            # this sorts tier-{2,3} builds below tier-1 in the index
+            # rank is one for non-tier-1 tasks and based on pushid for others;
+            # this sorts tier-{2,3} builds below tier-1 in the index, but above
+            # eager-index
             tier = task.get("treeherder", {}).get("tier", 3)
-            extra_index["rank"] = 0 if tier > 1 else int(config.params["build_date"])
+            extra_index["rank"] = 1 if tier > 1 else int(config.params["build_date"])
         elif rank == "build_date":
             extra_index["rank"] = int(config.params["build_date"])
         else:
@@ -1965,6 +2206,43 @@ def set_task_and_artifact_expiry(config, jobs):
         yield job
 
 
+def group_name_variant(group_names, groupSymbol):
+    # iterate through variants, allow for Base-[variant_list]
+    # sorting longest->shortest allows for finding variants when
+    # other variants have a suffix that is a subset
+    variant_symbols = sorted(
+        [
+            (
+                v,
+                TEST_VARIANTS[v]["suffix"],
+                TEST_VARIANTS[v].get("description", "{description}"),
+            )
+            for v in TEST_VARIANTS
+            if TEST_VARIANTS[v].get("suffix", "")
+        ],
+        key=lambda tup: len(tup[1]),
+        reverse=True,
+    )
+
+    # strip known variants
+    # build a list of known variants
+    base_symbol = groupSymbol
+    found_variants = []
+    for variant, suffix, description in variant_symbols:
+        if f"-{suffix}" in base_symbol:
+            base_symbol = base_symbol.replace(f"-{suffix}", "")
+            found_variants.append((variant, description))
+
+    if base_symbol not in group_names:
+        return ""
+
+    description = group_names[base_symbol]
+    for variant, desc in found_variants:
+        description = desc.format(description=description)
+
+    return description
+
+
 @transforms.add
 def build_task(config, tasks):
     for task in tasks:
@@ -2005,10 +2283,13 @@ def build_task(config, tasks):
             groupSymbol, symbol = split_symbol(task_th["symbol"])
             if groupSymbol != "?":
                 treeherder["groupSymbol"] = groupSymbol
-                if groupSymbol not in group_names:
+                description = group_names.get(
+                    groupSymbol, group_name_variant(group_names, groupSymbol)
+                )
+                if not description:
                     path = os.path.join(config.path, task.get("task-from", ""))
                     raise Exception(UNKNOWN_GROUP_NAME.format(groupSymbol, path))
-                treeherder["groupName"] = group_names[groupSymbol]
+                treeherder["groupName"] = description
             treeherder["symbol"] = symbol
             if len(symbol) > 25 or len(groupSymbol) > 25:
                 raise RuntimeError(
@@ -2048,6 +2329,8 @@ def build_task(config, tasks):
                 "kind": config.kind,
                 "label": task["label"],
                 "retrigger": "true" if attributes.get("retrigger", False) else "false",
+                "project": config.params["project"],
+                "trust-domain": config.graph_config["trust-domain"],
             }
         )
 
@@ -2072,6 +2355,8 @@ def build_task(config, tasks):
 
         if task.get("requires", None):
             task_def["requires"] = task["requires"]
+        if task.get("retries") is not None:
+            task_def["retries"] = task["retries"]
 
         if task_th:
             # link back to treeherder in description
@@ -2125,7 +2410,7 @@ def build_task(config, tasks):
             )
         attributes.setdefault("shipping_product", task["shipping-product"])
 
-        # Set MOZ_AUTOMATION on all jobs.
+        # Set some MOZ_* settings  on all jobs.
         if task["worker"]["implementation"] in (
             "generic-worker",
             "docker-worker",
@@ -2133,7 +2418,15 @@ def build_task(config, tasks):
             payload = task_def.get("payload")
             if payload:
                 env = payload.setdefault("env", {})
-                env["MOZ_AUTOMATION"] = "1"
+                env.update(
+                    {
+                        "MOZ_AUTOMATION": "1",
+                        "MOZ_BUILD_DATE": config.params["moz_build_date"],
+                        "MOZ_SCM_LEVEL": config.params["level"],
+                        "MOZ_SOURCE_CHANGESET": get_branch_rev(config),
+                        "MOZ_SOURCE_REPO": get_branch_repo(config),
+                    }
+                )
 
         dependencies = task.get("dependencies", {})
         if_dependencies = task.get("if-dependencies", [])
